"""Accumulate all the samples in samples/ for each experiment into a CSV"""
import numpy as np
import csv
from datetime import datetime
from os import listdir
import sys
from scipy import spatial

from make_graphs import plot
from make_graphs import sampling


def function_hull(v_x, v_y):
    """Linear hull of a non-negative function"""
    d_xmax = max(v_x)
    nda_graph = np.array([np.append(v_x, [0, d_xmax]),
                          np.append(v_y, [0, 0])]).T
    hull = spatial.ConvexHull(nda_graph)
    v_hull_idxs = np.sort(hull.vertices)

    if (v_hull_idxs[-1] == len(v_x)) | (v_hull_idxs[-1] == len(v_x)+1):
        v_hull_idxs = v_hull_idxs[:-1]
    if (v_hull_idxs[-1] == len(v_x)) | (v_hull_idxs[-1] == len(v_x)+1):
        v_hull_idxs = v_hull_idxs[:-1]

    v_hull = np.interp(v_x,
                       nda_graph[v_hull_idxs, 0],
                       nda_graph[v_hull_idxs, 1])
    return v_hull


def _accum_samples_bits(str_prefix):
    v_files = listdir('samples/')
    l_samples = list()
    for str_fname in v_files:
        if str_fname.startswith(str_prefix):
            sam = np.load('samples/'+str_fname).tolist()

            d_kfactor = sam['input']['v_kfactors'][0]
            d_inr = sam['input']['v_inrs'][0]
            v_max_bits = sam['input']['v_max_bits']
            n_trials = sam['input']['n_trials']

            # Update loaded sample (sam) to convex hulls of each series of bits
            n_pts = len(v_max_bits)
            v_cfgs = [(d_kfactor, d_inr, v_max_bits[n_pt]) for n_pt in range(n_pts)]
            for n_trial in range(n_trials):
                v_trial_dc_rates = [sam[t_cfg][n_trial].rate_opt_dc for t_cfg in v_cfgs]
                v_trial_ga_rates = [sam[t_cfg][n_trial].rate_opt_ga for t_cfg in v_cfgs]
                v_dc_hull_rates = function_hull(v_max_bits, v_trial_dc_rates)
                v_ga_hull_rates = function_hull(v_max_bits, v_trial_ga_rates)
                for iii in range(0, len(v_max_bits)):
                    sam[v_cfgs[iii]][n_trial].rate_opt_dc = v_dc_hull_rates[iii]
                    sam[v_cfgs[iii]][n_trial].rate_opt_ga = v_ga_hull_rates[iii]

            # Unroll sample into a list of observations
            l_sam = list()
            for key in sam.keys():
                if key == 'input':
                    continue
                l_key = [{
                    'k': key[0],
                    'inr': key[1],
                    'snr': 0,
                    'bits': key[2],
                    'ub': obs.rate_ub,
                    'bc': obs.rate_bc,
                    'ga': obs.rate_opt_ga,
                    'dc': obs.rate_opt_dc,
                    'v_dc_sorted': obs.v_dc_sorted
                          }
                    for obs in sam[key]]
                l_sam = l_sam + l_key
            l_samples = l_samples + l_sam
    return l_samples


def _accum_samples_k(str_prefix):
    v_files = listdir('samples/')
    l_samples = list()
    for str_fname in v_files:
        if str_fname.startswith(str_prefix):
            sam = np.load('samples/'+str_fname).tolist()

            # Unroll sample into a list of observations
            l_sam = list()
            for key in sam.keys():
                if key == 'input': continue
                l_key = [{
                    'k': key[0],
                    'inr': key[1],
                    'snr': 0,
                    'bits': key[2],
                    'ub': obs.rate_ub,
                    'bc': obs.rate_bc,
                    'ga': obs.rate_opt_ga,
                    'dc': obs.rate_opt_dc,
                    'v_dc_sorted': obs.v_dc_sorted
                }
                    for obs in sam[key]]
                l_sam = l_sam + l_key
            l_samples = l_samples + l_sam
    return l_samples


if __name__ == "__main__":
    startTime = datetime.now()

    str_prefix = 'k_ni_nb_'
    print(str_prefix)
    l_k_ni_nb = _accum_samples_k(str_prefix)
    for d in l_k_ni_nb:
        d.update({'base': False})

    str_prefix = 'k_ni_b_'
    print(str_prefix)
    l_k_ni_b = _accum_samples_k(str_prefix)
    for d in l_k_ni_b:
        d.update({'base': True})

    str_prefix = 'k_i_nb_'
    print(str_prefix)
    l_k_i_nb = _accum_samples_k(str_prefix)
    for d in l_k_i_nb:
        d.update({'base': False})

    str_prefix = 'k_i_b_'
    print(str_prefix)
    l_k_i_b = _accum_samples_k(str_prefix)
    for d in l_k_i_b:
         d.update({'base': True})

    l_keys = list(l_k_i_b[0].keys())
    with open('plot csv/all_vs_k.csv', 'w') as f:
        dict_writer = csv.DictWriter(f, l_keys)
        dict_writer.writeheader()
        dict_writer.writerows(l_k_ni_nb +
                   l_k_ni_b +
                   l_k_i_nb +
                   l_k_i_b)

    str_prefix = 'bits_ni_nb_'
    print(str_prefix)
    l_bits_ni_nb = _accum_samples_bits(str_prefix)
    for d in l_bits_ni_nb:
        d.update({'base': False})

    str_prefix = 'bits_ni_b_'
    print(str_prefix)
    l_bits_ni_b = _accum_samples_bits(str_prefix)
    for d in l_bits_ni_b:
        d.update({'base': True})

    str_prefix = 'bits_i_nb_'
    print(str_prefix)
    l_bits_i_nb = _accum_samples_bits(str_prefix)
    for d in l_bits_i_nb:
        d.update({'base': False})

    str_prefix = 'bits_i_b_'
    print(str_prefix)
    l_bits_i_b = _accum_samples_bits(str_prefix)
    for d in l_bits_i_b:
        d.update({'base': True})

    l_keys = list(l_bits_i_b[0].keys())
    with open('plot csv/all_vs_bits.csv', 'w') as f:
        dict_writer = csv.DictWriter(f, l_keys)
        dict_writer.writeheader()
        dict_writer.writerows(l_bits_ni_nb +
                   l_bits_ni_b +
                   l_bits_i_nb +
                   l_bits_i_b)

    print(datetime.now() - startTime)
    sys.exit(0)
