"""Check whether a distributed compression configuration violates entropy conditions"""
import numpy as np

D_MIN_USEFUL_BITRATE = 0.01


def checkdc(v_channel, nda_interf, v_dc_bits, v_ga_bits, d_lambda):
    """ Check whether a distributed compression configuration violates entropy conditions
        (continuous version)
         Input:
             v_channel = Channel vector h (size N+1 x 1)
             nda_interf = Interference covariance matrix Sigma (size N+1 x N+1)
             v_dc_bits = Post-compression helper encoding rates. (size N x 1)
             v_ga_bits = Pre-compression Gaussian encoder rates. (size N x 1)
             d_lambda = Hull parameter
         Output:
             b_is_sane = True only if v_dc_bits is valid given all the other inputs
     """
    v_badness = checkdc_cts(v_channel, nda_interf, v_dc_bits, v_ga_bits, d_lambda)
    b_is_sane = (max(v_badness) <= 0)
    return b_is_sane


def checkdc_cts(v_channel, nda_interf, v_dc_bits, v_ga_bits, d_lambda):
    """ Check whether a distributed compression configuration violates entropy conditions
        (continuous version)
         Input:
             v_channel = Channel vector h (size N+1 x 1)
             nda_interf = Interference covariance matrix Sigma (size N+1 x N+1)
             v_dc_bits = Post-compression helper encoding rates. (size N x 1)
             v_ga_bits = Pre-compression Gaussian encoder rates. (size N x 1)
             d_lambda = Hull parameter
         Output:
             v_badness = violation amounts
     """

    # Clip to useful receivers
    v_dc_bits_u = np.append(float('inf'), v_dc_bits)
    v_ga_bits_u = np.append(float('inf'), v_ga_bits)
    bv_using = (v_dc_bits_u > D_MIN_USEFUL_BITRATE) & (v_ga_bits_u > D_MIN_USEFUL_BITRATE)
    n_using = sum(bv_using)
    if n_using == 0:
        return 0

    v_channel_u = v_channel[bv_using]
    nda_interf_u = nda_interf[bv_using][:, bv_using]
    v_dc_bits_u = v_dc_bits_u[bv_using]
    v_ga_bits_u = v_ga_bits_u[bv_using]

    # Receiver thermal noise + interference covariance
    nda_receiver_noise = nda_interf_u + np.identity(n_using)

    # Receive signal covariance
    nda_receiver_signal = np.outer(v_channel_u, v_channel_u.conj())

    # Power of distortion due to quantization
    v_helper_distortion = np.real(np.diag(nda_receiver_signal + nda_receiver_noise))
    v_helper_distortion = np.divide(v_helper_distortion, np.power(2, v_ga_bits_u)-1)

    # Thermal + interference + distortion covariance
    nda_total_noise = nda_receiver_noise + np.diag(v_helper_distortion)

    """ 
    We want to compute H(Z_S|Z_{S^C}).
    ...but we only know:
     * H(Z_n) = v_ga_bits_u[n]
     * I(Z_S ; Z_{S^C}) = log2(|cov(Z_S)|*|cov(Z_{S^C})| / |cov(Z)|)

    Strategy:
    H(Z_S|Z_{S^C}) = H(Z_S) - I(Z_S;Z_{S^C})
    and:
    H(Z_S) = sum_{n in S} H(Z_n) 
             - sum_{all n} I(Z_sn ; Z_{sk, k<n})
    """

    v_s_n_dets = det_minors(nda_receiver_signal + nda_total_noise)
    v_cond_ents = np.zeros(2**n_using, dtype=np.complex64)

    v_badness = np.zeros(2**n_using, dtype=np.complex64)

    # skip empty set, since that is equivalent to the condition: lambda>0
    # skip sets that include the base; give vacuous conditions since the base has infinite bits
    for n_set in range(1, 2**(n_using-1)):
        n_set_c = 2**n_using - 1 - n_set

        # sum_{n in set} H(Z_n)
        bv_set = num2bv(n_set, n_using)
        d_entropy_include = sum(v_ga_bits_u[bv_set])

        # sum_{all n} I(Z_sn ; Z_{sk: k<n})
        d_entropy_exclude = 0
        n_in_set = sum(bv_set)
        n_set_kliii = 1
        for iii in range(1, n_in_set):
            n_set_iii = 2**iii
            n_set_kleqiii = n_set_kliii + n_set_iii
            d_entropy_exclude = d_entropy_exclude + \
                np.log2(v_s_n_dets[n_set_iii]*v_s_n_dets[n_set_kliii] /
                        v_s_n_dets[n_set_kleqiii])
            n_set_kliii = n_set_kleqiii

        # I(Z_S;Z_{S^C})
        d_mi = 0
        if not (n_set == 2**n_using - 1):
            d_mi = np.log2(v_s_n_dets[n_set]*v_s_n_dets[n_set_c] /
                           v_s_n_dets[-1])

        # entropy condition: v_badness[n_set] < 0
        d_cond_ent = d_entropy_include - d_entropy_exclude - d_mi
        v_cond_ents[n_set] = d_cond_ent
        v_badness[n_set] = np.real(d_cond_ent)-(sum(v_dc_bits_u[bv_set]) + d_lambda)

    return v_badness


def det_minors(nda_m):
    """ Compute the determinants of all the symmetric minors of a square matrix.

        Input:
            nda_m = N x N complex square np.ndarray
        Output:
            v_dets = vector of length 2^N where v_dets[iii] equals:
                            det( nda_m( bv_iii, bv_iii ) ),
                    where bv_iii is a binary vector representation of iii.
    """

    n_dim = nda_m.shape[0]
    v_dets = np.zeros(2**n_dim, dtype=np.complex64)
    for iii in range(1, 2**n_dim):
        bv_iii = num2bv(iii, n_dim)
        v_dets[iii] = np.linalg.det(nda_m[bv_iii][:, bv_iii])
    return v_dets


def num2bv(num, n_len):
    """Convert a number to a binary vector of some length"""
    return [bool((2**ii & num)//(2**ii)) for ii in range(n_len-1, -1, -1)]
