"""Optimize rates over a sum-bits constraint"""
from scipy import optimize
import numpy as np
from dist_recv import rates
from dist_recv import checkdc


def opt_ga(v_channel, nda_interf, d_max_bits):
    """ Optimize rates.rate_ga for a given channel, interference, over a sum-bits constraint
    Input:
        v_channel = Channel vector h (size N+1 x 1)
        nda_interf = Interference covariance matrix Sigma (size N+1 x N+1)
        d_max_bits = Max total helper encoding rates.
    Output:
        opt_result = Optimization output (v_bits)
    """
    n_helpers = len(v_channel)-1

    fn_objective = lambda v_bits: -rates.rate_ga(v_channel, nda_interf, v_bits)
    v_intl_guess = 1*np.ones(n_helpers)*d_max_bits/n_helpers
    nda_bounds = np.tile((0, None), (n_helpers, 1))
    fn_sum_constraint = lambda v_bits: d_max_bits - sum(v_bits)
    dict_constraints = dict(type='ineq', fun=fn_sum_constraint)
    opt_result = optimize.minimize(fn_objective,
                                   v_intl_guess,
                                   bounds=nda_bounds,
                                   constraints=dict_constraints)
    return opt_result


def rate_opt_ga(v_channel, nda_interf, opt_result_ga):
    """Given an opt_ga optimization result, return the associated ga rate it achieves"""
    v_bits = opt_result_ga.x
    return rates.rate_ga(v_channel, nda_interf, v_bits)


def opt_dc(v_channel, nda_interf, d_max_bits):
    """ 
    Optimize rates.rate_dc for a given channel, interference, over a sum-bits constraint
        (barrier relaxation)
    Input:
        v_channel = Channel vector h (size N+1 x 1)
        nda_interf = Interference covariance matrix Sigma (size N+1 x N+1)
        d_max_bits = Max total helper encoding rates.
    Output:
        opt_result = Optimization output:
                     opt_result.x[0:N] = v_dc_bits
                     opt_result.x[N:-1] = v_ga_bits
                     opt_result.x[-1] = d_lambda
    """
    n_helpers = len(v_channel)-1

    v_opt_ga = opt_ga(v_channel, nda_interf, d_max_bits).x
    d_opt_ga = rates.rate_ga(v_channel, nda_interf, v_opt_ga)

    v_intl_dc = np.ones(n_helpers)*d_max_bits/n_helpers
    v_intl_ga = v_intl_dc
    d_intl_lambda = 0
    v_best = np.append([v_intl_dc, v_intl_ga], np.array([d_intl_lambda]))

    nda_bounds = np.tile([0, d_max_bits], (n_helpers*2+1, 1))

    fn_sum_bd = lambda v_config: d_max_bits - sum(v_config[0:n_helpers])
    constraint_sum = {'type': 'ineq', 'fun': fn_sum_bd}

    fn_objective = lambda v_config: fn_dc_objective(v_channel, nda_interf, v_config)

    opt_result = None
    d_brlx = -2
    while True:
        fn_dc_bd = lambda v_config: -max(checkdc.checkdc_cts(v_channel,
                                                             nda_interf,
                                                             v_dc_bits=v_config[0:n_helpers],
                                                             v_ga_bits=v_config[n_helpers:-1],
                                                             d_lambda=v_config[-1])-d_brlx)
        constraint_dc = {'type': 'ineq', 'fun': fn_dc_bd}
        v_constraints = [constraint_sum, constraint_dc]

        opt_result = optimize.minimize(fn_objective,
                                       v_best,
                                       bounds=nda_bounds,
                                       constraints=v_constraints)

        # Only update the initial guess if it's better than GA
        b_is_worse = rate_opt_dc(v_channel, nda_interf, opt_result) < d_opt_ga
        if (sum(np.isnan(opt_result.x)) > 0) | (b_is_worse & (d_brlx >= -0.05)):
            opt_result.x = np.append([v_intl_dc, v_intl_ga], np.array([d_intl_lambda]))
        else:
            v_intl_dc = opt_result.x[0:n_helpers]
            v_intl_ga = opt_result.x[n_helpers:-1]
            d_intl_lambda = opt_result.x[-1]
            v_best = np.append([v_intl_dc, v_intl_ga], np.array([d_intl_lambda]))

        d_brlx = d_brlx/4
        if d_brlx > -0.02:
            break

    # Give up
    b_is_worse = rate_opt_dc(v_channel, nda_interf, opt_result) < d_opt_ga
    if (sum(np.isnan(opt_result.x)) > 0) | b_is_worse:
        opt_result.x = np.append([v_opt_ga, v_opt_ga], np.array([0]))
        # print('Optimization failed; using GA')

    return opt_result


def rate_opt_dc(v_channel, nda_interf, opt_result_dc):
    """Given an opt_dc optimization result, return the associated dc rate it achieves"""
    n_helpers = len(v_channel)-1
    v_config = opt_result_dc.x
    return rates.rate_dc(v_channel=v_channel,
                         nda_interf=nda_interf,
                         v_dc_bits=v_config[0:n_helpers],
                         v_ga_bits=v_config[n_helpers:-1],
                         d_lambda=v_config[-1])


def fn_dc_objective(v_channel, nda_interf, v_config):
    if(sum(np.isnan(v_config))) > 0:
        return float('inf')
    n_helpers = v_channel.shape[0]-1
    d_rate = rates.rate_ga(v_channel, nda_interf, v_bits=v_config[n_helpers:-1])
    d_dc_obj = -(d_rate - v_config[-1])
    return d_dc_obj
