"""Rates for Distributed Receive"""
import numpy as np

from dist_recv import checkdc as dc

D_MIN_USEFUL_BITRATE = 0.01


def rate_ub(v_channel, nda_interf, d_bits):
    """ Upper bound on rate for distributed receive with reception at base
        Input:
            v_channel = Channel vector h (size N+1 x 1)
            nda_interf = Interference covariance matrix Sigma (size N+1 x N+1)
            d_bits = Total amount of helper bits
        Output:
            d_rate = Rate (in bits)
    """
    n_recv = v_channel.shape[0]

    # Receiver thermal noise + interference covariance
    nda_receiver_noise = nda_interf + np.identity(n_recv)

    # Receive signal covariance
    nda_receiver_signal = np.outer(v_channel, v_channel.conj())

    d_rate = np.log2(np.real(np.linalg.det(nda_receiver_signal + nda_receiver_noise)) /
                     np.real(np.linalg.det(nda_receiver_noise)))

    d_rate_0 = np.log2(np.real(1+nda_interf[0, 0] + np.abs(v_channel[0])**2) /
                       np.real(1+nda_interf[0, 0]))
    if not (d_rate_0 > 0):
        d_rate_0 = 0

    d_rate = min(d_rate, d_bits + d_rate_0)
    return d_rate


def rate_bc(v_channel, nda_interf, d_bits):
    """ Rate for Broadcast distributed receive with reception at base
        Input:
            v_channel = Channel vector h (size N+1 x 1)
            nda_interf = Interference covariance matrix Sigma (size N+1 x N+1)
            d_bits = Total number of bits used by helpers
        Output:
            d_rate = Rate (in bits)
    """
    d_rate_0 = np.log2(np.real(1+nda_interf[0, 0] + np.abs(v_channel[0])**2) /
                       np.real(1+nda_interf[0, 0]))
    if not (d_rate_0 > 0):
        d_rate_0 = 0
    idx_n = np.argmax(np.abs(v_channel))
    d_rate_n = np.log2(np.real(1+nda_interf[idx_n, idx_n] + np.abs(v_channel[idx_n])**2) /
                       np.real(1+nda_interf[idx_n, idx_n]))
    if not (d_rate_n > 0):
        d_rate_n = 0
    d_rate_bc = max([d_rate_0, min([d_rate_n, d_bits])])
    return d_rate_bc


def rate_ga(v_channel, nda_interf, v_bits):
    """ Rate for Gaussian approximation to distributed receive with reception at base
        Input:
            v_channel = Channel vector h (size N+1 x 1)
            nda_interf = Interference covariance matrix Sigma (size N+1 x N+1)
            v_bits = Helper encoding rates. (size N x 1)
        Output:
            d_rate = Rate (in bits)
    """

    # Give up if there are any nans
    if sum(np.isnan(v_bits)) > 0:
        return 0

    # Clip to useful receivers
    v_bits_u = np.append(float('inf'), v_bits)
    v_using = v_bits_u > D_MIN_USEFUL_BITRATE
    n_using = sum(v_using)
    if n_using == 0:
        return 0

    v_channel_u = v_channel[v_using]
    nda_interf_u = nda_interf[v_using][:, v_using]
    v_bits_u = v_bits_u[v_using]

    # Receiver thermal noise + interference covariance
    nda_receiver_noise = nda_interf_u + np.identity(n_using)

    # Receive signal covariance
    nda_receiver_signal = np.outer(v_channel_u, v_channel_u.conj())

    # Power of distortion due to quantization
    v_helper_distortion = np.real(np.diag(nda_receiver_signal + nda_receiver_noise))
    v_helper_distortion = np.divide(v_helper_distortion, np.power(2, v_bits_u)-1)

    # Thermal + interference + distortion covariance
    nda_total_noise = nda_receiver_noise + np.diag(v_helper_distortion)

    d_rate = np.log2(np.real(np.linalg.det(nda_receiver_signal + nda_total_noise)) /
                     np.real(np.linalg.det(nda_total_noise)))

    if not (np.isreal(d_rate)):
        raise ValueError('rate_ga gives complex rate for arguments')
    if d_rate < 0:
        raise ValueError('rate_ga gives negative rate for arguments')

    return d_rate


def rate_dc(v_channel, nda_interf, v_dc_bits, v_ga_bits, d_lambda):
    """ Rate for Distributed Compression distributed receive with reception at base
        Input:
            v_channel = Channel vector h (size N+1 x 1)
            nda_interf = Interference covariance matrix Sigma (size N+1 x N+1)
            v_dc_bits = Post-compression helper encoding rates. (size N x 1)
            v_ga_bits = Pre-compression Gaussian encoder rates. (size N x 1)
            d_lambda = Hull parameter
        Output:
            d_rate = Rate (in bits), 0 if configuration is not valid
    """
    b_is_valid = dc.checkdc(v_channel, nda_interf, v_dc_bits, v_ga_bits, d_lambda)
    if b_is_valid:
        return rate_ga(v_channel, nda_interf, v_ga_bits) - d_lambda
    else:
        return 0

def rate_dcu(v_channel, nda_interf, d_bits):
    """ Rate for Distributed Compression distributed receive with reception at base
        Input:
            v_channel = Channel vector h (size N+1 x 1)
            nda_interf = Interference covariance matrix Sigma (size N+1 x N+1)
            d_bits = Total number of bits to evenly distribute.
        Output:
            d_rate = Rate (in bits), 0 if configuration is not valid
    """

    n_helpers = size(v_channel) - 1
    v_dc_bits = np.ones(n_helpers)*n_helpers/d_bits
    v_ga_bits = v_dc_bits
    d_step_size = 1.0

    while d_step_size > 0.001
        v_step = np.random.rand(n_helpers)
        v_step = v_step/sum(v_step)*d_step_size
        v_ga_new = v_ga_bits + v_step

        b_is_valid = dc.checkdc(v_channel, nda_interf, v_dc_bits, v_ga_new, 0)
        if ~b_is_valid:
            d_step_size = d_step_size*0.9
            continue
        else:
            d_step_size = d_step_size/0.9
            v_ga_bits = v_ga_new

    return rate_ga(v_channel, nda_interf, v_ga_bits/1.01)
