"""Maximize bounds over trials of a random 0dB-average-SNR channel"""
import numpy as np
import multiprocessing
import collections
import itertools

from dist_recv import rates, opt_rates


def random_sample(n_recv, v_kfactors, v_inrs, v_max_bits, n_trials, b_no_base=False, n_detail_lvl=0):
    """ Run random trials with the specified parameters 
    Input:
        n_recv = number of receivers
        v_kfactors = Channel K-factors (dB) to run trials on (unique elements)
        v_inr = INRs (dB) to run trials on (unique elements)
        v_max_bits = sum-bitrates to run trials on  (unique elements)
        n_trials = Number of trials to run for each combination of params in v_kfactors, v_inr, v_lanbits
        b_no_base = True if base doesn't have its own reception
    Output:
        dict_sample.<input> = <input>
            dict_sample['input'] = named tuple (n_recv, v_kfactors, v_inr, v_lanbits, n_trials) 
            dict_sample.(t_params) = l_trials (t_params = tuple in itertools.product(v_kfactors, v_inr, v_lanbits))
                l_trials[i] = DistRecvTrial containing all the info about the i-th trial (i from 0 to n_trials-1)
    """
    dict_sample = collections.defaultdict()
    dict_sample['input'] = {'n_recv': n_recv,
                            'v_kfactors': v_kfactors,
                            'v_inrs': v_inrs,
                            'v_max_bits': v_max_bits,
                            'n_trials': n_trials,
                            'b_no_base': b_no_base}

    for n_trial in range(n_trials):
        if n_detail_lvl >= 1:
            print('TRIAL ' + repr(n_trial + 1) + ' ***********************')

        v_channel_unscaled = (np.random.standard_normal(n_recv) +
                              1j*np.random.standard_normal(n_recv))/np.sqrt(2)
        v_interf_angles = np.exp(1j*2*np.pi*np.random.uniform(0, 1, n_recv))

        if b_no_base:
            v_channel_unscaled[0] = 0
            v_interf_angles[0] = 0

        nda_interf_unscaled = np.outer(v_interf_angles, v_interf_angles.conj().T)

        v_params = itertools.product(v_kfactors,
                                     v_inrs,
                                     v_max_bits,
                                     [n_recv],
                                     [v_channel_unscaled],
                                     [nda_interf_unscaled],
                                     [n_detail_lvl])
        with multiprocessing.Pool(16) as pool:
            for t_params, drt in pool.map(_process_key, v_params):
            # for t_params, drt in map(_process_key, v_params):
                if t_params in dict_sample:
                    dict_sample[t_params].append(drt)
                else:
                    dict_sample[t_params] = [drt]

    return dict_sample


def _process_key(t_params_aug):
    t_params = t_params_aug[0:3]
    d_kfactor_db = t_params[0]
    d_inr_db = t_params[1]
    d_max_bits = t_params[2]

    n_recv = t_params_aug[3]
    v_channel_unscaled = t_params_aug[4]
    nda_interf_unscaled = t_params_aug[5]
    n_detail_lvl = t_params_aug[6]

    d_kfactor = np.power(10, d_kfactor_db / 10)
    d_inr = np.power(10, d_inr_db / 10)

    # average receive signal power (i.e. linear SNR)
    d_omega = 1
    d_direct_power = (d_kfactor / (d_kfactor + 1)) * d_omega
    d_scattered_power = d_omega / (d_kfactor + 1)
    v_channel = (np.sqrt(d_direct_power) * np.exp(1j * 2 * np.pi * np.random.uniform(0, 1, n_recv)) +
                 np.sqrt(d_scattered_power) * v_channel_unscaled)

    nda_interf = np.sqrt(d_inr) * nda_interf_unscaled

    if n_detail_lvl >= 2:
        print('   K = ' + repr(t_params[0]) + ' dB\n' +
              ' INR = ' + repr(t_params[1]) + ' dB\n' +
              'bits = ' + repr(t_params[2]))

    drt = DistRecvTrial(v_channel, nda_interf, d_max_bits)

    return t_params, drt


def sample_average(dict_sample, dict_avg=dict()):
    """ Given an output from random_sample, return the sample average over trials
        Input:
        dict_sample = sample to average
        dict_avg = dictionary to update (default is an empty dict)
        Output:
        dict_avg
        dict_avg['nt_input'] = named tuple, sample input
    """
    dict_avg['input'] = dict_sample['input']

    n_trials = dict_avg['input']['n_trials']
    v_kfactors = dict_avg['input']['v_kfactors']
    v_inrs = dict_avg['input']['v_inrs']
    v_max_bits = dict_avg['input']['v_max_bits']

    if not ('n_trials' in dict_avg):
        dict_avg['n_trials'] = 0

    n_params = len(v_kfactors)*len(v_inrs)*len(v_max_bits)
    for t_params in itertools.product(v_kfactors, v_inrs, v_max_bits):
        for n_t in range(n_trials):
            drt = dict_sample[t_params][n_t]
            if t_params in dict_avg:
                dict_avg[t_params].update(drt)
            else:
                dict_avg[t_params] = SampleAverage(drt)
            dict_avg['n_trials'] = dict_avg['n_trials'] + 1/n_params
    dict_avg['n_trials'] = np.round(dict_avg['n_trials'])
    return dict_avg


class DistRecvTrial:
    """ A set of rates and information corresponding to a specific distributed receive scenario
    """
    def __init__(self, v_channel, nda_interf, d_max_bits):
        n_recv = len(v_channel)

        self.v_channel = v_channel
        self.nda_interf = nda_interf
        self.d_max_bits = d_max_bits

        self.rate_ub = rates.rate_ub(v_channel, nda_interf, d_max_bits)
        self.rate_bc = rates.rate_bc(v_channel, nda_interf, d_max_bits)

        self.opt_ga = opt_rates.opt_ga(v_channel, nda_interf, d_max_bits)
        self.rate_opt_ga = opt_rates.rate_opt_ga(v_channel, nda_interf, self.opt_ga)

        self.opt_dc = opt_rates.opt_dc(v_channel, nda_interf, d_max_bits)
        self.rate_opt_dc = opt_rates.rate_opt_dc(v_channel, nda_interf, self.opt_dc)
        self.v_dc_sorted = np.sort(self.opt_dc.x[0:(n_recv-1)])


class SampleAverage:
    """ A set of rates and info corresponding to averages over random_sample.DistRecvTrial's """

    def __init__(self, drt):
        self.v_channel = drt.v_channel
        self.nda_interf = drt.nda_interf
        self.d_max_bits = drt.d_max_bits
        self.rate_ub = drt.rate_ub
        self.rate_bc = drt.rate_bc
        self.rate_opt_ga = drt.rate_opt_ga
        self.rate_opt_dc = drt.rate_opt_dc
        self.v_dc_sorted = drt.v_dc_sorted

        self.n_trials = 1

    def update(self, drt):
        """Update the running average rates given a new DistRecvTrial"""
        n_t = self.n_trials
        self.v_channel = (n_t * self.v_channel + drt.v_channel) / (n_t + 1)
        self.nda_interf = (n_t * self.nda_interf + drt.nda_interf) / (n_t + 1)
        self.d_max_bits = (n_t * self.d_max_bits + drt.d_max_bits) / (n_t + 1)
        self.rate_ub = (n_t * self.rate_ub + drt.rate_ub) / (n_t + 1)
        self.rate_bc = (n_t * self.rate_bc + drt.rate_bc) / (n_t + 1)
        self.rate_opt_ga = (n_t * self.rate_opt_ga + drt.rate_opt_ga) / (n_t + 1)
        self.rate_opt_dc = (n_t * self.rate_opt_dc + drt.rate_opt_dc) / (n_t + 1)
        self.v_dc_sorted = (n_t * self.v_dc_sorted + drt.v_dc_sorted) / (n_t + 1)

        self.n_trials = n_t + 1
