"""Test rate function for Gaussian approximation to distributed receive"""
import numpy as np

from dist_recv import rates


def test_ga_zero_channel():
    """Make sure rate_ga with a zero channel is zero"""
    n_recv = 5
    v_channel = np.zeros(n_recv)
    nda_interf = 1j*np.random.standard_normal((n_recv, n_recv)) + \
                    np.random.standard_normal((n_recv, n_recv))
    nda_interf = np.dot(nda_interf, nda_interf.conj().T)/2
    v_bits = np.random.standard_normal(n_recv-1)**2
    assert rates.rate_ga(v_channel, nda_interf, v_bits) < 10e-4


def test_ga_no_base_reception():
    """Make sure rate_ga with no base observation is less than with base observation"""
    n_recv = 5
    v_channel = 1j*np.random.standard_normal(n_recv) + np.random.standard_normal(n_recv)
    v_channel = v_channel*1/np.sqrt(2)
    nda_interf = 1j*np.random.standard_normal((n_recv, n_recv)) + \
                    np.random.standard_normal((n_recv, n_recv))
    nda_interf = np.dot(nda_interf, nda_interf.conj().T)/2
    v_bits = np.random.standard_normal(n_recv-1)**2
    d_rate_with_base = rates.rate_ga(v_channel, nda_interf, v_bits)
    v_channel[0] = 0
    nda_interf[0,:] = np.zeros(n_recv)
    nda_interf[:,0] = np.zeros(n_recv)
    d_rate_without_base = rates.rate_ga(v_channel, nda_interf, v_bits)
    d_diff = d_rate_with_base - d_rate_without_base
    assert d_diff > 0


def test_bc_leq_ub():
    """Make sure rate_ga is less than rate_ub"""
    n_recv = 5
    v_channel = 1j * np.random.standard_normal(n_recv) + np.random.standard_normal(n_recv)
    v_channel = v_channel * 1 / np.sqrt(2)
    nda_interf = 1j * np.random.standard_normal((n_recv, n_recv)) + \
                 np.random.standard_normal((n_recv, n_recv))
    nda_interf = np.dot(nda_interf, nda_interf.conj().T) / 2
    d_bits = np.random.standard_normal(1)**2
    d_rate_ub = rates.rate_ub(v_channel, nda_interf, d_bits)
    d_rate_bc = rates.rate_bc(v_channel, nda_interf, d_bits)
    d_diff = d_rate_ub - d_rate_bc
    assert d_diff >= 0


def test_ga_leq_ub():
    """Make sure rate_ga is less than rate_ub"""
    n_recv = 5
    v_channel = 1j * np.random.standard_normal(n_recv) + np.random.standard_normal(n_recv)
    v_channel = v_channel * 1 / np.sqrt(2)
    nda_interf = 1j * np.random.standard_normal((n_recv, n_recv)) + \
                 np.random.standard_normal((n_recv, n_recv))
    nda_interf = np.dot(nda_interf, nda_interf.conj().T) / 2
    v_bits = np.random.standard_normal(n_recv-1)**2
    d_bits = sum(v_bits)
    d_rate_ga = rates.rate_ga(v_channel, nda_interf, v_bits)
    d_rate_ub = rates.rate_ub(v_channel, nda_interf, d_bits)
    d_diff = d_rate_ub - d_rate_ga
    assert d_diff >= 0


def test_ga_leq_dc():
    """Make sure rate_ga is less than rate_dc with v_bits_dc = v_bits_ga 
    in particular, this should be a valid config. """
    n_recv = 5
    v_channel = 1j * np.random.standard_normal(n_recv) + np.random.standard_normal(n_recv)
    v_channel = v_channel * 1 / np.sqrt(2)
    nda_interf = 1j * np.random.standard_normal((n_recv, n_recv)) + \
                 np.random.standard_normal((n_recv, n_recv))
    nda_interf = np.dot(nda_interf, nda_interf.conj().T) / 2
    v_bits = 5 * np.ones(n_recv - 1)
    d_lambda = 0;
    d_rate_ga = rates.rate_ga(v_channel, nda_interf, v_bits)
    d_rate_dc = rates.rate_dc(v_channel, nda_interf, v_bits, v_bits, d_lambda)
    d_diff = d_rate_dc - d_rate_ga
    assert d_diff >= 0
